#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifndef USE_NETLIB
#include <net/ethernet.h>
#include <netinet/in.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>
#else
#include <netlib.h>
#endif

#include "pktlib.h"
#include "pktbuf.h"

struct dhcp {
  unsigned char message_type;
  unsigned char hardware_type;
  unsigned char hardware_addrlen;
  unsigned char hops;
  unsigned int transaction_id;
  unsigned short seconds_elapsed;
  unsigned short flags;
  unsigned int client_ip_address;
  unsigned int your_ip_address;
  unsigned int next_server_ip_address;
  unsigned int relay_agent_ip_address;
  char client_mac_address[6];
  unsigned char padding[10];
  unsigned char server_host_name[64];
  unsigned char boot_file_name[128];
  unsigned int magic_cookie;
#define DHCP_MAGIC_COOKIE_DHCP 0x63825363
  unsigned char option[0];
};

struct dhcp_option {
  unsigned char type;
#define DHCP_OPTION_TYPE_SUBNET_MASK   0x01
#define DHCP_OPTION_TYPE_HOST_NAME     0x0c
#define DHCP_OPTION_TYPE_DOMAIN_NAME   0x0f
#define DHCP_OPTION_TYPE_REQ_IP_ADDR   0x32
#define DHCP_OPTION_TYPE_IP_LEASE_TIME 0x33
#define DHCP_OPTION_TYPE_MESSAGE_TYPE  0x35
#define DHCP_OPTION_TYPE_SERVER_ID     0x36
#define DHCP_OPTION_TYPE_PARAM_REQLIST 0x37
#define DHCP_OPTION_TYPE_CLIENT_ID     0x3d
#define DHCP_OPTION_TYPE_END           0xff
  unsigned char length;
  unsigned char data[0];
};

#define DHCP_OPTION_MESSAGE_TYPE_NONE     0
#define DHCP_OPTION_MESSAGE_TYPE_DISCOVER 1
#define DHCP_OPTION_MESSAGE_TYPE_OFFER    2
#define DHCP_OPTION_MESSAGE_TYPE_REQUEST  3
#define DHCP_OPTION_MESSAGE_TYPE_ACK      5
#define DHCP_OPTION_MESSAGE_TYPE_RELEASE  7

struct server {
  struct ether_addr mac_address;
  unsigned int ip_address;
  unsigned int subnet_mask;
};

struct client {
  struct ether_addr mac_address;
  unsigned int ip_address;
  int status;
  unsigned int transaction_id;
};

static struct client *search_client(int clients_num, struct client *clients,
				    char *macaddr)
{
  int i;
  char zero[6] = { 0, 0, 0, 0, 0, 0 };

  if (macaddr == NULL)
    macaddr = zero;

  for (i = 0; i < clients_num; i++) {
    if (!memcmp(&clients[i].mac_address, macaddr, 6))
      return &clients[i];
  }

  return NULL;
}

static char *macaddr_s(char macaddr[])
{
  return ether_ntoa((struct ether_addr *)macaddr);
}

static char *ipaddr_s(unsigned int ipaddr)
{
  struct in_addr in;
  in.s_addr = htonl(ipaddr);
  return inet_ntoa(in);
}

static char *type2name(int type)
{
  switch (type) {
  case DHCP_OPTION_MESSAGE_TYPE_NONE:     return "NONE";
  case DHCP_OPTION_MESSAGE_TYPE_DISCOVER: return "DISCOVER";
  case DHCP_OPTION_MESSAGE_TYPE_OFFER:    return "OFFER";
  case DHCP_OPTION_MESSAGE_TYPE_REQUEST:  return "REQUEST";
  case DHCP_OPTION_MESSAGE_TYPE_ACK:      return "ACK";
  case DHCP_OPTION_MESSAGE_TYPE_RELEASE:  return "RELEASE";
  default: break;
  }
  return "UNKNOWN";
}

static int send_dhcp_offer_or_ack(pktif_t pktif, struct header_param *paramp,
				  struct server *server, struct client *client,
				  int type)
{
  pktbuf_t pktbuf;
  struct header_param param;
  struct dhcp *dhcphdr;
  struct dhcp_option *dhcpopt;
  unsigned int val4;

  memcpy(param.ethernet.src_macaddr, &server->mac_address, ETHER_ADDR_LEN);
  memcpy(param.ethernet.dst_macaddr, &paramp->ethernet.src_macaddr, ETHER_ADDR_LEN);
  param.ethernet.type = ETHERTYPE_IP;

  param.ip.id = 0x1234;
  param.ip.ttl = 64;
  param.ip.protocol = IPPROTO_UDP;
  param.ip.src_ipaddr = server->ip_address;
  param.ip.dst_ipaddr = client->ip_address;

  param.udp.src_port = 67;
  param.udp.dst_port = 68;

  pktbuf = pktbuf_create(-1);
  dhcphdr = (struct dhcp *)pktbuf_add_header(pktbuf, sizeof(*dhcphdr) + 64);
  memset(dhcphdr, 0, sizeof(*dhcphdr));
  dhcphdr->message_type = 2; /* Boot Reply */
  dhcphdr->hardware_type = 1; /* Ethernet */
  dhcphdr->hardware_addrlen = ETHER_ADDR_LEN;
  dhcphdr->transaction_id = htonl(client->transaction_id);
  dhcphdr->your_ip_address = htonl(client->ip_address);
  memcpy(dhcphdr->client_mac_address, &client->mac_address, ETHER_ADDR_LEN);
  dhcphdr->magic_cookie = htonl(DHCP_MAGIC_COOKIE_DHCP);
  dhcpopt = (struct dhcp_option *)dhcphdr->option;

  dhcpopt->type = DHCP_OPTION_TYPE_MESSAGE_TYPE;
  dhcpopt->length = 1;
  dhcpopt->data[0] = type;
  dhcpopt = (struct dhcp_option *)((char *)dhcpopt + 2 + dhcpopt->length);

  dhcpopt->type = DHCP_OPTION_TYPE_SERVER_ID;
  dhcpopt->length = 4;
  val4 = htonl(server->ip_address);
  memcpy(dhcpopt->data, &val4, sizeof(val4));
  dhcpopt = (struct dhcp_option *)((char *)dhcpopt + 2 + dhcpopt->length);

  dhcpopt->type = DHCP_OPTION_TYPE_IP_LEASE_TIME;
  dhcpopt->length = 4;
  val4 = htonl(600);
  memcpy(dhcpopt->data, &val4, sizeof(val4));
  dhcpopt = (struct dhcp_option *)((char *)dhcpopt + 2 + dhcpopt->length);

  dhcpopt->type = DHCP_OPTION_TYPE_SUBNET_MASK;
  dhcpopt->length = 4;
  val4 = htonl(server->subnet_mask);
  memcpy(dhcpopt->data, &val4, sizeof(val4));
  dhcpopt = (struct dhcp_option *)((char *)dhcpopt + 2 + dhcpopt->length);

  dhcpopt->type = DHCP_OPTION_TYPE_DOMAIN_NAME;
  dhcpopt->length = 11;
  memcpy(dhcpopt->data, "example.org", 11);
  dhcpopt = (struct dhcp_option *)((char *)dhcpopt + 2 + dhcpopt->length);

  dhcpopt->type = DHCP_OPTION_TYPE_END;
  dhcpopt = (struct dhcp_option *)((char *)dhcpopt + 1);

  pktbuf_set_size(pktbuf, (char *)dhcpopt - (char *)dhcphdr);

  pktbuf = pktbuf_make_udp(pktbuf, &param);
  pktbuf = pktbuf_make_ip(pktbuf, &param);
  pktbuf = pktbuf_checksum_correct_ip(pktbuf);
  pktbuf = pktbuf_make_ethernet(pktbuf, &param);

  pktbuf = pktbuf_send(pktif, pktbuf);
  if (pktbuf) {
    pktbuf_destroy(pktbuf);
    return 0;
  }

  return 1;
}

static int analyze_dhcp(struct dhcp *dhcphdr, struct server *server,
			struct client *client)
{
  int type = DHCP_OPTION_MESSAGE_TYPE_NONE;
  struct dhcp_option *dhcpopt;
  unsigned int ipaddr;
  int mismatch = 0;

  if (dhcphdr->message_type != 1) /* Boot Request */
    goto err;
  if (dhcphdr->hardware_type != 1) /* Ethernet */
    goto err;
  if (dhcphdr->hardware_addrlen != ETHER_ADDR_LEN)
    goto err;
  if (ntohl(dhcphdr->magic_cookie) != DHCP_MAGIC_COOKIE_DHCP)
    goto err;

  client->transaction_id = ntohl(dhcphdr->transaction_id);
  memcpy(&client->mac_address, dhcphdr->client_mac_address, ETHER_ADDR_LEN);

  printf("Client MAC Address : %s\n", macaddr_s(dhcphdr->client_mac_address));

  dhcpopt = (struct dhcp_option *)dhcphdr->option;

  while (1) {
    if (dhcpopt->type == DHCP_OPTION_TYPE_END)
      break;
    if (dhcpopt->type == DHCP_OPTION_TYPE_MESSAGE_TYPE) {
      type = dhcpopt->data[0];
      printf("Message Type        : %s\n", type2name(type));
    }
    if (dhcpopt->type == DHCP_OPTION_TYPE_SERVER_ID) {
      ipaddr = htonl(server->ip_address);
      if (memcmp(&ipaddr, dhcpopt->data, sizeof(ipaddr)))
	mismatch++;
      memcpy(&ipaddr, dhcpopt->data, sizeof(ipaddr));
      ipaddr = ntohl(ipaddr);
      printf("Server ID           : %s\n", ipaddr_s(ipaddr));
    }
    if (dhcpopt->type == DHCP_OPTION_TYPE_REQ_IP_ADDR) {
      ipaddr = htonl(client->ip_address);
      if (memcmp(&ipaddr, dhcpopt->data, sizeof(ipaddr)))
	mismatch++;
      memcpy(&ipaddr, dhcpopt->data, sizeof(ipaddr));
      ipaddr = ntohl(ipaddr);
      printf("Requested IP Address: %s\n", ipaddr_s(ipaddr));
    }
    dhcpopt = (struct dhcp_option *)((char *)dhcpopt + 2 + dhcpopt->length);
  }

  if (mismatch && (type != DHCP_OPTION_MESSAGE_TYPE_DISCOVER))
    goto err;

  return type;

err:
  return DHCP_OPTION_MESSAGE_TYPE_NONE;
}

static int proc_dhcp(pktif_t pktif, struct header_param *param,
		     struct dhcp *dhcphdr, struct server *server,
		     int clients_num, struct client *clients)
{
  int type;
  struct client *client;

  client = search_client(clients_num, clients, dhcphdr->client_mac_address);
  if (client == NULL)
    client = search_client(clients_num, clients, NULL);
  if (client == NULL)
    return -1;

  type = analyze_dhcp(dhcphdr, server, client);

  printf("Old Status          : %s\n", type2name(client->status));
  switch (type) {
  case DHCP_OPTION_MESSAGE_TYPE_DISCOVER:
    send_dhcp_offer_or_ack(pktif, param, server, client,
			   DHCP_OPTION_MESSAGE_TYPE_OFFER);
    client->status = DHCP_OPTION_MESSAGE_TYPE_OFFER;
    break;
  case DHCP_OPTION_MESSAGE_TYPE_REQUEST:
    send_dhcp_offer_or_ack(pktif, param, server, client,
			   DHCP_OPTION_MESSAGE_TYPE_ACK);
    client->status = DHCP_OPTION_MESSAGE_TYPE_ACK;
    break;
  case DHCP_OPTION_MESSAGE_TYPE_RELEASE:
    client->status = DHCP_OPTION_MESSAGE_TYPE_NONE;
    break;
  default:
    break;
  }
  printf("New Status          : %s\n", type2name(client->status));

  return type;
}

static int recv_dhcp(pktif_t pktif, struct server *server,
		     int clients_num, struct client *clients)
{
  pktbuf_t pktbuf;
  int r = 0, next, payload = 0;
  struct header_param param;

  pktbuf = pktbuf_recv(pktif, 0);
  if (pktbuf)
    pktbuf_destroy(pktbuf);

  next = pktif_get_linktype(pktif);

  switch (next) {
  case DLT_EN10MB:
    r = pktbuf_analyze_ethernet(pktbuf, &param);
    break;
  default:
    break;
  }

  if (r > 0) {
    switch (next) {
    case DLT_EN10MB:
      printf("Ethernet:\n");
      printf("\tsrc macaddr: %s\n", macaddr_s(param.ethernet.src_macaddr));
      printf("\tdst macaddr: %s\n", macaddr_s(param.ethernet.dst_macaddr));
      printf("\ttype       : %04x\n", param.ethernet.type);
      payload += param.ethernet.header_size;
      next = param.ethernet.type;
      break;
    default:
      next = -1;
      break;
    }
  }

  if (r > 1) {
    switch (next) {
    case ETHERTYPE_IP:
      printf("IP:\n");
      printf("\tid        : %04x\n", param.ip.id);
      printf("\tttl       : %04x\n", param.ip.ttl);
      printf("\tprotocol  : %d\n", param.ip.protocol);
      printf("\tsrc ipaddr: %s\n", ipaddr_s(param.ip.src_ipaddr));
      printf("\tdst ipaddr: %s\n", ipaddr_s(param.ip.dst_ipaddr));
      payload += param.ip.header_size;
      next = param.ip.protocol;
      break;
    default:
      next = -1;
      break;
    }
  }

  if (r > 2) {
    switch (next) {
    case IPPROTO_UDP:
      printf("UDP:\n");
      printf("\tsrc port: %d\n", param.udp.src_port);
      printf("\tdst port: %d\n", param.udp.dst_port);
      payload += param.udp.header_size;
      if ((param.udp.src_port == 68) && (param.udp.dst_port == 67)) {
	return proc_dhcp(pktif, &param,
			 (struct dhcp *)(pktbuf_get_header(pktbuf) + payload),
			 server, clients_num, clients);
      }
      break;
    default:
      break;
    }
  }

  return 0;
}

int main(int argc, char *argv[])
{
  pktif_t pktif;
  struct server server;
  struct client *clients;
  int clients_num, i;

  argc--;
  argv++;

  pktif = pktif_open(argv[0], PKTIF_OPEN_FLAG_RECV_NOTSENT, 0);

  memcpy(&server.mac_address, ether_aton(argv[1]), ETHER_ADDR_LEN);
  server.ip_address = (unsigned int)ntohl(inet_addr(argv[2]));
  server.subnet_mask = (unsigned int)ntohl(inet_addr(argv[3]));

  clients_num = atoi(argv[4]);
  clients = malloc(sizeof(*clients) * clients_num);

  for (i = 0; i < clients_num; i++) {
    memset(&clients[i], 0, sizeof(*clients));
    clients[i].ip_address = (unsigned int)ntohl(inet_addr(argv[5])) + i;
    clients[i].status = DHCP_OPTION_MESSAGE_TYPE_NONE;
  }

  while (1) {
    if (recv_dhcp(pktif, &server, clients_num, clients) < 0)
      break;
  }

  return 0;
}
